//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKSteppingAction.hh"
#include "VKEventAction.hh"
#include "VKTrackingAction.hh"
#include "VKTrajectory.hh"
#include "VKSiPMSensitiveDetector.hh"
#include "VKUserTrackInformation.hh"
#include "VKUserEventInformation.hh"
#include "VKSteppingMessenger.hh"
#include "VKRecorderBase.hh"

#include "VKScintillatorSensitiveDetector.hh"

#include "G4SteppingManager.hh"
#include "G4SDManager.hh"
#include "G4EventManager.hh"
#include "G4ProcessManager.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4Event.hh"
#include "G4StepPoint.hh"
#include "G4TrackStatus.hh"
#include "G4VPhysicalVolume.hh"
#include "G4ParticleDefinition.hh"
#include "G4ParticleTypes.hh"


VKSteppingAction::VKSteppingAction(VKRecorderBase* r)
 : fRecorder(r),
   fOneStepPrimaries(false)
{
    fSteppingMessenger = new VKSteppingMessenger(this);
    fExpectedNextStatus = Undefined;
}


VKSteppingAction::~VKSteppingAction() 
{
    // do nothing
}


void VKSteppingAction::UserSteppingAction(const G4Step * theStep)
{
    // track
    G4Track* theTrack = theStep -> GetTrack();
    if (theTrack->GetCurrentStepNumber() == 1) fExpectedNextStatus = Undefined;
 
    // track info
    VKUserTrackInformation* trackInfo
        = (VKUserTrackInformation*)theTrack -> GetUserInformation();

    // event info
    VKUserEventInformation* eventInfo
        = (VKUserEventInformation*)G4EventManager::GetEventManager()
            -> GetConstCurrentEvent() -> GetUserInformation();

    // pre-step info
    G4StepPoint* thePrePoint = theStep -> GetPreStepPoint();
    G4VPhysicalVolume* thePrePV = thePrePoint -> GetPhysicalVolume();
    
    // post-step info
    G4StepPoint* thePostPoint = theStep -> GetPostStepPoint();
    G4VPhysicalVolume* thePostPV = thePostPoint -> GetPhysicalVolume();

    // boundary process
    G4OpBoundaryProcessStatus boundaryStatus = Undefined;
    static G4ThreadLocal G4OpBoundaryProcess* boundary = NULL;

    // find the boundary process
    if (!boundary) {
        // process manager
        G4ProcessManager* pm = theTrack -> GetDefinition() -> GetProcessManager();
        // number of processes
        G4int nprocesses = pm -> GetProcessListLength();
        // list of processes
        G4ProcessVector* pv = pm->GetProcessList();
        // search for boundary process
        for (G4int i=0; i<nprocesses; i++) {
            if( (*pv)[i]->GetProcessName() == "OpBoundary") {
                boundary = (G4OpBoundaryProcess*)(*pv)[i];
                break; // stop after first boundary process has been found
            }
        }
    }
    
    // If this is a primary track
    if (theTrack->GetParentID() == 0) {
 
        // secondaries
        G4TrackVector* fSecondary = fpSteppingManager -> GetfSecondary();

        // total number of secondaries
        G4int tN2ndariesTot = fpSteppingManager -> GetfN2ndariesAtRestDoIt()
                            + fpSteppingManager -> GetfN2ndariesAlongStepDoIt()
                            + fpSteppingManager -> GetfN2ndariesPostStepDoIt();

        // If we havent already found the conversion position and there were
        // secondaries generated, then search for it
        if (!eventInfo->IsConvPosSet() && tN2ndariesTot>0 ) {
            size_t lp1min = (*fSecondary).size()-tN2ndariesTot; 
            size_t lp1max = (*fSecondary).size();
            // loop over secondaries
            for (size_t lp1=lp1min; lp1<lp1max; lp1++) {
                // get process that created secondary
                const G4VProcess* creator=(*fSecondary)[lp1]->GetCreatorProcess();
                if (creator) {
                    G4String creatorName=creator->GetProcessName();
                    // if the process is any of the following, store the position
                    if (creatorName=="phot"||creatorName=="compt"||creatorName=="conv") {
                        // since this is happening before the secondary is being tracked
                        // the Vertex position has not been set yet(set in initial step)
                        eventInfo -> SetConvPos((*fSecondary)[lp1]->GetPosition());
                    }
                }
            }
        }

        // If user has selected to only allow primaries to go one step
        // before being killed, kill track now
        if (fOneStepPrimaries && (thePrePV->GetName()=="signal" || thePrePV->GetName()=="veto")) {
            theTrack -> SetTrackStatus(fStopAndKill);
        }
    }
    
    // if particle is exiting the world
    if (!thePostPV) {
        fExpectedNextStatus=Undefined;
        return;
    }

    // If particle is an optical photon
    G4ParticleDefinition* particleType = theTrack->GetDefinition();
    if (particleType == G4OpticalPhoton::OpticalPhotonDefinition()) {

        // kill optical photons entering `world`
        if (thePostPV->GetName() == "world") theTrack -> SetTrackStatus(fStopAndKill);

        // Check which SiPM/Scintillator the optical photon is in
        auto preName   = thePrePV->GetName(); 
        auto preCopyNo = thePrePV->GetCopyNo();
        G4int SiPMScintId = 0;
        if (preName == "signal" || (preName == "SiPM" && preCopyNo == 0))    SiPMScintId = 0;
        else if (preName == "veto" || (preName == "SiPM" && preCopyNo == 1)) SiPMScintId = 1;

        // If photon was absorbed by the absorption process
        G4String procName = thePostPoint->GetProcessDefinedStep()->GetProcessName();
        if (procName == "OpAbsorption") {
            eventInfo -> IncAbsorption(SiPMScintId);
            trackInfo -> AddTrackStatusFlag(absorbed);
        }

        boundaryStatus = boundary -> GetStatus();        

        // Check to see if the partcile was actually at a boundary
        // Otherwise the boundary status may not be valid
        if (thePostPoint->GetStepStatus() == fGeomBoundary) {
            if (fExpectedNextStatus == StepTooSmall) {
                if (boundaryStatus != StepTooSmall) {
                    G4ExceptionDescription ed;
                    ed << "VKSteppingAction::UserSteppingAction(): "
                       << "No reallocation step after reflection!"
                       << G4endl;
                    G4Exception("VKSteppingAction::UserSteppingAction()", "VKExpl01",
                        FatalException, ed, "Something is wrong with the surface normal or geometry");
                }
            }
            fExpectedNextStatus = Undefined;
            
            // Depending on boundary status, take different course of action
            switch (boundaryStatus) {
                case Absorption:
                    trackInfo->AddTrackStatusFlag(boundaryAbsorbed);
                    eventInfo->IncBoundaryAbsorption(SiPMScintId);
                    break;
                case Detection: {
                    //
                    // Note, this assumes that the volume causing detection
                    // is the photocathode because it is the only one with
                    // non-zero efficiency.
                    // Trigger sensitive detector manually since photon is
                    // absorbed and status is Detection
                    //
                    G4SDManager* SDman = G4SDManager::GetSDMpointer();
                    G4String sdName = "SiPM_SD";
                    VKSiPMSensitiveDetector* sipmSD = (VKSiPMSensitiveDetector*)SDman -> FindSensitiveDetector(sdName);
                    if (sipmSD) sipmSD -> ProcessHits_constStep(theStep, NULL);
                    trackInfo -> AddTrackStatusFlag(hitSiPM);
                    break;
                }
                case FresnelReflection:
                case TotalInternalReflection:
                case LambertianReflection:
                case LobeReflection:
                case SpikeReflection:
                case BackScattering:
                    trackInfo -> IncReflections(); // this gets executed for all types of reflections
                    fExpectedNextStatus = StepTooSmall;
                    break;
                default:
                    break;
            }
        }
    } // end if is an optical photon

    if (fRecorder) fRecorder -> RecordStep(theStep);
    
    // kill track if step number exceeds 10,000
    if (theTrack->GetCurrentStepNumber() > 1E4) theTrack->SetTrackStatus(fStopAndKill);
}

/*
 *  1: Undefined 	
 *  2: FresnelRefraction 	
 *  3: FresnelReflection 	
 *  4: TotalInternalReflection 	
 *  5: LambertianReflection 	
 *  6: LobeReflection 	
 *  7: SpikeReflection 	
 *  8: BackScattering 	
 *  9: Absorption 	
 * 10: Detection 	
 * 11: NotAtBoundary 	
 * 12: SameMaterial 	
 * 13: StepTooSmall 	
 * 14: NoRINDEX 
*/
